#include <bits/libc-lock.h>
#include <errno.h>
#include <pgsql/libpq-fe.h>
#include <nss.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "nss_setup.h"

/* Locks the static variables in this file.  */
static pthread_mutex_t _nss_pg_lock;

/* global var : the connection handler */
static PGconn *pgconn = NULL;


static enum nss_status cursor_begin(void) { 
   char query[BUFFER]="";		/* query string */
   PGresult *pgres;				/* handler for postgresql answers */

#if DEBUG>2
   _nss_pgsql_log(LOG_DEBUG,"beginning cursor");
#endif
   snprintf(query,BUFFER,
	   "DECLARE pwcursor BINARY CURSOR FOR "
	   "select %s,%s,%s,%s,%s,%s,%s from %s",
	   PW_NAME_FIELD,PW_PASSWD_FIELD,PW_UID_FIELD,
	   PW_GID_FIELD,PW_GECOS_FIELD,PW_DIR_FIELD,
	   PW_SHELL_FIELD,USER_TABLE);
   pgres=PQexec(pgconn,query);
   if (!pgres || PQresultStatus(pgres) != PGRES_COMMAND_OK) {
#if DEBUG>0
	   _nss_pgsql_log(LOG_ERR,"%s",PQerrorMessage(pgconn));
#endif
	   PQclear(pgres);
	   return NSS_STATUS_UNAVAIL;
   }
   PQclear(pgres);
   return NSS_STATUS_SUCCESS;
}

static void cursor_close(void) { 
   PGresult *pgres;				/* handler for postgresql answers */
#if DEBUG>2
   _nss_pgsql_log(LOG_DEBUG,"closing cursor");
#endif
   if (pgconn != NULL) { 
      pgres=PQexec(pgconn,"CLOSE pwcursor");
      PQclear(pgres);
      pgres=PQexec(pgconn,"COMMIT");
      PQclear(pgres);
   }
}

static enum nss_status internal_setpwent (void)
{

   if (pgconn && PQstatus(pgconn)==CONNECTION_BAD) { 
#if DEBUG>2
         _nss_pgsql_log(LOG_DEBUG,"resetting connection");
#endif
         PQreset(pgconn);
	}

   if (!pgconn) {
      char *conn_strings = NULL;	/* postgresql strings of connections options */
      struct stat st;				/* stat info of config file */
      FILE *conf;						/* handler of config file */
      PGresult *pgres;				/* handler for postgresql answers */
      char **server_list;			/* array of conenction strings */
      char *bindex;
      int lindex;
      int whatsleft;
      int linesize;

#if DEBUG>2
      _nss_pgsql_log(LOG_DEBUG,"starting connection");
#endif
      if (stat(CONF_FILE,&st)) { /* check if config file exixt */
         return NSS_STATUS_UNAVAIL;
      }
      /* try to allocate mem for whole config file */
      if (!(conn_strings = malloc(st.st_size+MAXSERVERS+1))) { 
         return NSS_STATUS_UNAVAIL;
      }
      if (!(conf = fopen(CONF_FILE,"r"))) { /* try open it */
         free(conn_strings);
         return NSS_STATUS_UNAVAIL;
      }
      /* try to allocate mem for the array of servers */
      if (!(server_list = malloc((sizeof (char *))* (MAXSERVERS + 1)))) {
         free(conn_strings);
	      fclose(conf);
         return NSS_STATUS_UNAVAIL;
      }
      
      bindex = conn_strings;
      lindex = 0;
      whatsleft = st.st_size;
      fgets(bindex,whatsleft,conf);
      while (!feof(conf) && lindex != MAXSERVERS) {
         server_list[lindex++] = bindex;
         linesize = strlen(bindex)-1;
	      bindex[linesize]='\0'; /* delete trailing newline */
         whatsleft -= linesize;
         bindex += linesize;
         fgets(bindex,whatsleft,conf);
      }
      server_list[lindex] = NULL;
      while (*server_list[--lindex]=='\0') { 
			server_list[lindex] = NULL;
		} /* trailing newline*/
#if DEBUG>5		
      _nss_pgsql_log(LOG_DEBUG,"find %d servers definition.",lindex);
#endif      
      fclose(conf);
      lindex = 0;
      while (!pgconn && server_list[lindex]) {
#if DEBUG>4      
         _nss_pgsql_log(LOG_DEBUG,"connect string :[%d] %s",lindex,server_list[lindex]);
#endif
         pgconn=PQconnectdb(server_list[lindex++]);
         if ((pgconn == NULL) || (PQstatus(pgconn) == CONNECTION_BAD)) {
	         PQfinish(pgconn);
	         pgconn = NULL;
		 	}
      }
      
      free(conn_strings);
      free(server_list);

      if ((pgconn == NULL) || (PQstatus(pgconn) == CONNECTION_BAD)) { 
#if DEBUG>0
         _nss_pgsql_log(LOG_ERR,"PgERROR %s",PQerrorMessage(pgconn));
#endif
         if (pgconn!=NULL) PQfinish(pgconn);
         pgconn=NULL;
         return NSS_STATUS_UNAVAIL;
      } else {
#if DEBUG>3
         _nss_pgsql_log(LOG_DEBUG,"connected");
#endif 
			pgres=PQexec(pgconn,"BEGIN");
			if (!pgres || PQresultStatus(pgres) != PGRES_COMMAND_OK)
			{
#if DEBUG>0
				_nss_pgsql_log(LOG_ERR,"%s",PQerrorMessage(pgconn));
#endif
				PQclear(pgres);
				PQfinish(pgconn);
				pgconn=NULL;
				return NSS_STATUS_UNAVAIL;
			}
			PQclear(pgres);
      } 
   } else { 
#if DEBUG>3
      _nss_pgsql_log(LOG_DEBUG,"already connected");
#endif
   }
  
   return NSS_STATUS_SUCCESS;
}


static void internal_endpwent (void)
{

#if DEBUG>2
   _nss_pgsql_log(LOG_DEBUG,"closing connection");
#endif

   if (pgconn!=NULL) {
      PQfinish(pgconn);
      pgconn = NULL;
#if DEBUG>2
      _nss_pgsql_log(LOG_DEBUG,"connection closed");
#endif
   }
}


enum nss_status fill_record (struct passwd *pwd, char *buffer, 
	size_t bufflen, PGresult *pgres)
{
   enum nss_status status=NSS_STATUS_SUCCESS;
	int remain=bufflen;
   char *tmp,*index=buffer;

	tmp=PQgetvalue(pgres,0,0);
	remain-=strlen(tmp)+1;
	if (remain > 0 ) { 
		pwd->pw_name=strcpy(index,tmp);
		index+=strlen(index)+1;
	} else { 
		status = NSS_STATUS_TRYAGAIN;
	}
	tmp=PQgetvalue(pgres,0,1);
	remain-=strlen(tmp)+1;
	if (remain > 0 ) { 
		pwd->pw_passwd=strcpy(index,tmp);
		index+=strlen(index)+1;
	} else { 
		status = NSS_STATUS_TRYAGAIN;
	}
	pwd->pw_uid=atoi(PQgetvalue(pgres,0,2));
	pwd->pw_gid=atoi(PQgetvalue(pgres,0,3));
	tmp=PQgetvalue(pgres,0,4);
	remain-=strlen(tmp)+1;
	if (remain > 0 ) { 
		pwd->pw_gecos=strcpy(index,tmp);
		index+=strlen(index)+1;
	} else { 
		status = NSS_STATUS_TRYAGAIN;
	}
	tmp=PQgetvalue(pgres,0,5);
	remain-=strlen(tmp)+1;
	if (remain > 0 ) { 
		pwd->pw_dir=strcpy(index,tmp);
		index+=strlen(index)+1;
	} else { 
		status = NSS_STATUS_TRYAGAIN;
	}
	tmp=PQgetvalue(pgres,0,6);
	remain-=strlen(tmp)+1;
	if (remain > 0 ) { 
		pwd->pw_shell=strcpy(index,tmp);
		index+=strlen(index)+1;
	} else { 
		status = NSS_STATUS_TRYAGAIN;
	}
	return status;
}


static enum nss_status lookup (const char *name, const char *type, 
   struct passwd *pwd, char *buffer, size_t bufflen, int *errnop) {

  enum nss_status status;
  PGresult *pgres = NULL;
  char query[BUFFER]="";
  

   if (!strncmp("uid",type,3)) { 
#if DEBUG>2
    _nss_pgsql_log(LOG_DEBUG,"getpwuid(%s)",name);
#endif
    snprintf(query,BUFFER,"select %s,%s,%s,%s,%s,%s,%s from %s where %s = %s",
      PW_NAME_FIELD,PW_PASSWD_FIELD,PW_UID_FIELD,PW_GID_FIELD,PW_GECOS_FIELD,PW_DIR_FIELD,
      PW_SHELL_FIELD,USER_TABLE,PW_UID_FIELD,name);
  }
  
  if (!strncmp("nam",type,3)) {
#if DEBUG>2
    _nss_pgsql_log(LOG_DEBUG,"getpwnam(%s)",name);
#endif
    snprintf(query,BUFFER,"select %s,%s,%s,%s,%s,%s,%s from %s where %s = '%s'",
      PW_NAME_FIELD,PW_PASSWD_FIELD,PW_UID_FIELD,PW_GID_FIELD,PW_GECOS_FIELD,PW_DIR_FIELD,
      PW_SHELL_FIELD,USER_TABLE,PW_NAME_FIELD,name);
  }
  
   if (!strncmp("ent",type,3)) {
#if DEBUG>2
      _nss_pgsql_log(LOG_DEBUG,"getpwent()");
#endif
      snprintf(query,BUFFER,"FETCH NEXT from pwcursor");
   }

  if (strlen(query)==0) { 
     return NSS_STATUS_NOTFOUND;
  }
  
#if DEBUG>4
  _nss_pgsql_log(LOG_DEBUG,"%s",query);
#endif
  
  pgres=PQexec(pgconn,query);

  if (PQresultStatus(pgres) == PGRES_TUPLES_OK ) {
    if ( PQntuples(pgres) == 1 ) { 
#if DEBUG>3
		_nss_pgsql_log(LOG_DEBUG,"one entry found",query);
#endif
      status = fill_record(pwd,buffer,bufflen,pgres);
    } else { 
      /* zero or more than one entry found */
#if DEBUG>1
       _nss_pgsql_log(LOG_WARNING,"%d entry found",PQntuples(pgres));
#endif
      status = NSS_STATUS_NOTFOUND;
    } 
  } else { 
    /* query errors */
#if DEBUG>0
    _nss_pgsql_log(LOG_ERR,"%s",PQresultErrorMessage(pgres));
#endif 
    status = NSS_STATUS_NOTFOUND;
  }
  
  PQclear(pgres);
  
  return status;
}

enum nss_status _nss_pgsql_setpwent (void) {
   enum nss_status status;

   __libc_lock_lock (_nss_pg_lock);
   status = internal_setpwent ();
   if (status == NSS_STATUS_SUCCESS) {
      status = cursor_begin();
   }
   __libc_lock_unlock (_nss_pg_lock);
   return status;
}

enum nss_status _nss_pgsql_endpwent (void) {

   __libc_lock_lock (_nss_pg_lock);
   cursor_close();
   internal_endpwent();
   __libc_lock_unlock (_nss_pg_lock);
   return NSS_STATUS_SUCCESS;
}

enum nss_status _nss_pgsql_getpwnam_r (const char *name, 
   struct passwd *pwd, char *buffer, size_t buflen, int *errnop) {
  enum nss_status status;

  __libc_lock_lock (_nss_pg_lock);
  status = internal_setpwent ();
  if (status == NSS_STATUS_SUCCESS) {
     status = lookup (name, "nam", pwd, buffer, buflen, errnop);
     internal_endpwent();
  }
#if DEBUG>2  
  _nss_pgsql_log(LOG_DEBUG,"getpwname(%s) return [%d]",name,status);
#endif
  __libc_lock_unlock (_nss_pg_lock);
  return status;
}

enum nss_status _nss_pgsql_getpwuid_r (uid_t uid, struct passwd *pwd,
			char *buffer, size_t buflen, int *errnop) {
  enum nss_status status;
  char uidstr[21];	/* We will probably never have a gid_t with more than 64 bit. */

  snprintf (uidstr, sizeof uidstr, "%d", uid);
  __libc_lock_lock (_nss_pg_lock);
  status = internal_setpwent ();
  if (status == NSS_STATUS_SUCCESS) {
		status = lookup (uidstr, "uid", pwd, buffer, buflen, errnop);
		internal_endpwent();
  }
#if DEBUG>2  
  _nss_pgsql_log(LOG_DEBUG,"getpwuid(%d) return [%d]",uid,status);
#endif  
  __libc_lock_unlock (_nss_pg_lock);
  return status;
}

enum nss_status _nss_pgsql_getpwent_r (struct passwd *pwd,
		char *buffer, size_t buflen, int *errnop)
{
	enum nss_status status;

	__libc_lock_lock (_nss_pg_lock);
	status = lookup (NULL, "ent", pwd, buffer, buflen, errnop);
	__libc_lock_unlock (_nss_pg_lock);
	return status;
}

